#include "pch.h"

#include "notifier.h"
#include "resource.h"

#define FONT_HEIGHT     20
#define WINDOW_HEIGHT   32
#define MAX_STR_CHARS   32

typedef struct _CAPTION
{
    UINT    cch;
    WCHAR   szText[MAX_STR_CHARS];
}
CAPTION;

static const PCWSTR _szClassName = L"Caps Lock Notifier";
static HFONT _hFont;
static HPEN _hPen;
static HBRUSH _hBrush;
static int _nWindowWidth;
static CAPTION _captionCapsLockOn, _captionCapsLockOff;
static CAPTION * _pCaption;
static HWND _hWnd;

static void PaintDC(HDC hDC)
{
    SelectObject(hDC, _hPen);
    SelectObject(hDC, _hBrush);
    Rectangle(hDC, 1, 1, _nWindowWidth, WINDOW_HEIGHT);
    SetTextColor(hDC, RGB(0, 0, 0));
    SetTextAlign(hDC, TA_CENTER | TA_BOTTOM);
    SetBkMode(hDC, TRANSPARENT);
    SelectObject(hDC, _hFont);
    ExtTextOutW(
        hDC,
        _nWindowWidth / 2,
        (WINDOW_HEIGHT + FONT_HEIGHT) / 2,
        0,
        NULL,
        _pCaption->szText,
        _pCaption->cch,
        NULL
        );
}

static void PaintWindow(HWND hWnd)
{
    HDC hDCWindow, hDCMem;
    PAINTSTRUCT paint;
    const RECT * pRect;
    HBITMAP hBitmap;

    hDCWindow = BeginPaint(hWnd, &paint);
    // Cannot continue if paint.rcPaint is not initialized with valid data.
    if (hDCWindow == NULL) return;
    pRect = &paint.rcPaint;
    hDCMem = CreateCompatibleDC(hDCWindow);
    hBitmap =
        CreateCompatibleBitmap(
        hDCWindow,
        pRect->right - pRect->left,
        pRect->bottom - pRect->top
        );
    SelectObject(hDCMem, hBitmap);
    SetWindowOrgEx(hDCMem, pRect->left, pRect->top, NULL);
    PaintDC(hDCMem);
    BitBlt(
        hDCWindow,
        pRect->left,
        pRect->top,
        pRect->right - pRect->left,
        pRect->bottom - pRect->top,
        hDCMem,
        pRect->left,
        pRect->top,
        SRCCOPY
        );
    DeleteDC(hDCMem);
    DeleteObject(hBitmap);
    EndPaint(hWnd, &paint);
}

static LRESULT CALLBACK
    WndProc(HWND hWnd, UINT message, WPARAM wparam, LPARAM lparam)
{
    switch (message)
    {
    case WM_CREATE:
        _hWnd = hWnd;
        break;
    case WM_DESTROY:
        _hWnd = NULL;
        break;
    case WM_PAINT:
        PaintWindow(hWnd);
        break;
    case WM_TIMER:
        DestroyWindow(hWnd);
        break;
    default:
        return DefWindowProc(hWnd, message, wparam, lparam);
    }
    return 0;
}

static int LoadCaption(UINT uID, CAPTION * pCaption)
{
    return
        pCaption->cch = LoadStringW(NULL, uID, pCaption->szText, MAX_STR_CHARS);
}

static int GetCaptionWidth(HDC hDC, const CAPTION * pCaption)
{
    SIZE size;

    return
        GetTextExtentPoint32W(hDC, pCaption->szText, pCaption->cch, &size) ?
        size.cx : 0;
}

// Initialize the captions before calling this function.
static int GetCaptionMaxWidthDC(HDC hDC)
{
    int nWidth1, nWidth2;

    nWidth1 = GetCaptionWidth(hDC, &_captionCapsLockOn);
    if (nWidth1 == 0) return 0;
    nWidth2 = GetCaptionWidth(hDC, &_captionCapsLockOff);
    if (nWidth2 == 0) return 0;
    return nWidth1 > nWidth2 ? nWidth1 : nWidth2;
}

// Initialize font and captions before calling this function.
static int GetCaptionMaxWidth()
{
    HDC hDC;
    int nWidth;
    
    hDC = GetDC(NULL);
    nWidth = SelectObject(hDC, _hFont) ? GetCaptionMaxWidthDC(hDC) : 0;
    ReleaseDC(NULL, hDC);

    return nWidth;
}

BOOL InitCapsLockNotifier()
{
    WNDCLASSW wndClass;
    HFONT hFont;
    HPEN hPen;
    int nWidth;
    
    wndClass.style          =
        CS_HREDRAW | CS_VREDRAW | CS_SAVEBITS | CS_DROPSHADOW;
    wndClass.lpfnWndProc    = WndProc;
    wndClass.cbClsExtra     = 0;
    wndClass.cbWndExtra     = 0;
    wndClass.hInstance      = NULL;
    wndClass.hIcon          = NULL;
    wndClass.hCursor        = LoadCursorA(NULL, IDC_ARROW);
    wndClass.hbrBackground  = NULL;
    wndClass.lpszMenuName   = NULL;
    wndClass.lpszClassName  = _szClassName;
    if (!RegisterClassW(&wndClass)) return FALSE;

    hFont =
        CreateFontW(
        20,
        0,
        0,
        0,
        FW_BOLD,
        FALSE,
        FALSE,
        FALSE,
        ANSI_CHARSET,
        OUT_DEFAULT_PRECIS,
        CLIP_DEFAULT_PRECIS,
        PROOF_QUALITY,
        VARIABLE_PITCH | FF_MODERN,
        NULL
        );
    if (hFont == NULL) return FALSE;
    _hFont = hFont;

    hPen = CreatePen(PS_SOLID, 2, RGB(0, 0, 0));
    if (hPen == NULL) return FALSE;
    _hPen = hPen;

    _hBrush = (HBRUSH)GetStockObject(WHITE_BRUSH);
    
    if (
        LoadCaption(IDS_CAPS_LOCK_ON,   &_captionCapsLockOn)    == 0 ||
        LoadCaption(IDS_CAPS_LOCK_OFF,  &_captionCapsLockOff)   == 0)
        return FALSE;

    nWidth = GetCaptionMaxWidth();
    if (nWidth == 0) return FALSE;
    _nWindowWidth = nWidth + 36;

    return TRUE;
}

static int Min0(int n)
{
    return n > 0 ? n : 0;
}

void ShowCapsLockNotifier()
{
    RECT rect;

    if (_hWnd != NULL) DestroyWindow(_hWnd);
    if (SystemParametersInfoW(SPI_GETWORKAREA, 0, &rect, 0))
    {
        HWND hWnd;
        
        const int nWidth    = _nWindowWidth;
        const int nHeight   = WINDOW_HEIGHT;
        const int nMargin   = 16;
        hWnd =
            CreateWindowExW(
            WS_EX_TOPMOST | WS_EX_NOACTIVATE,
            _szClassName,
            NULL,
            WS_POPUP,
            Min0(rect.right - nWidth - nMargin),
            Min0(rect.bottom - nHeight - nMargin),
            nWidth,
            nHeight,
            NULL,
            NULL,
            NULL,
            NULL
            );
        _pCaption =
            GetKeyState(VK_CAPITAL) & 0x01 ?
            &_captionCapsLockOn : &_captionCapsLockOff;
        ShowWindow(hWnd, SW_SHOWNA);
        SetTimer(hWnd, 0, 3000, NULL);
    }
}
